package seu.agent.fb.FireBrigadestools;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Comparator;
import java.util.List;
import java.util.Set;
import java.util.SortedSet;
import java.util.TreeSet;

import javolution.util.FastSet;
import rescuecore2.misc.Pair;
import rescuecore2.standard.entities.Area;
import rescuecore2.standard.entities.Building;
import rescuecore2.standard.entities.FireBrigade;
import rescuecore2.standard.entities.Human;
import rescuecore2.standard.entities.StandardEntity;
import rescuecore2.worldmodel.EntityID;
import seu.Viewer.SelectedObject;
import seu.agent.fb.FireBrigadestools.simulators.WaterCoolingEstimator;
import seu.geom.Distance;
import seu.model.SEUWorldModel;
import seu.model.AgentConstants;
import seu.model.object.SEUBuilding;
import seu.model.route.pov.POVRouter;


/**
 * This class offer some utility tools for FB.
 *
 */
public class FbUtilities {
	private static final int EXTINGUISH_DISTANCE_THRESHOLD = 5000;

	private SEUWorldModel world;
	private POVRouter router;
	
	// constructor
	public FbUtilities(SEUWorldModel world) {
		this.world = world;
		this.router = world.getRouter();
	}
	
	/**
	 * Calculate the total needed water to extinguish the target building
	 * 
	 * @param building
	 *            the target building
	 * @return the total needed water of the target building
	 */
	public static int waterNeededToExtinguish(SEUBuilding building) {
		int groundArea = building.getSelfBuilding().getGroundArea();
    	int floors = building.getSelfBuilding().getFloors();
    	int buildingCode = building.getSelfBuilding().getBuildingCode();
    	double temperature = building.getEstimatedTemperature();
    	
    	return WaterCoolingEstimator.getWaterNeeded(groundArea, floors, buildingCode, temperature, 20);
	}
	
	/**
	 * Calculate the water power of an Agent to extinguish the target building.
	 * 
	 * @param world
	 *            the world model
	 * @param building
	 *            the target building
	 * @return the water power FB Agent will used to extinguish the target
	 *         building
	 */
	public static int calculateWaterPower(SEUWorldModel world, SEUBuilding building) {
		int agentWater = ((FireBrigade) world.getControlledEntity()).getWater();
		int maxPower = world.getConfig().maxPower;

		return Math.min(agentWater, maxPower);
	}

	/**
	 * Rerank buildings to increase the building value of unreachable buildings.
	 * 
	 * @param buildings
	 *            a set of old ranked building
	 * @param fbAgent
	 *            FB Agent
	 * @return the reranked building set
	 */
	public SortedSet<Pair<Pair<EntityID, Double>, Double>> reRankBuildings(
			SortedSet<Pair<Pair<EntityID, Double>, Double>> buildings, FireBrigade fbAgent) {
		SEUBuilding seuBuilding;
		EntityID agentId = fbAgent.getID();
		SortedSet<Pair<Pair<EntityID, Double>, Double>> result = new TreeSet<Pair<Pair<EntityID, Double>, Double>>(pairComparator_new);
		Set<SEUBuilding> buildingsExtinguishable = getBuildingInExtinguishableRange(this.world, agentId);
		boolean inMyExtiguishableRange;

		/*
		 * For building that has higher BUILDING_VALUE has small possibility in
		 * extinguishable distance, so we do not need to consider its
		 * reachability and just put it in is ok.
		 */
		int i = 0, AstarCount = 10;
		if (world.isMapMedium())
			AstarCount = 5;
		if (world.isMapHuge())
			AstarCount = 3;

		for (Pair<Pair<EntityID, Double>, Double> next : buildings) {
			seuBuilding = world.getSeuBuilding(next.first().first());

			if (i >= AstarCount) {
				result.add(new Pair<Pair<EntityID, Double>, Double>(next.first(), seuBuilding.BUILDING_VALUE));
				i++;
				continue;
			}

			inMyExtiguishableRange = buildingsExtinguishable.contains(seuBuilding);
			EntityID location = getNearest(world,seuBuilding.getAreasInExtinguishableRange(), agentId);

			if (inMyExtiguishableRange && isReacable(world, router, location)) {
				result.add(new Pair<Pair<EntityID, Double>, Double>(next.first(), seuBuilding.BUILDING_VALUE));
				i++;
			} else {
				seuBuilding.BUILDING_VALUE *= 10000;
				result.add(new Pair<Pair<EntityID, Double>, Double>(next.first(), seuBuilding.BUILDING_VALUE));
				i++;
			}
		}

		if (AgentConstants.LAUNCH_VIEWER) {
			SelectedObject.renderBuildingValueKey = true;
		}
		
		return result;
	}
	
	/**
	 * Get all buildings within the extinguishable range of the source Agent.
	 * 
	 * @param world
	 *            the world model
	 * @param source
	 *            the source Agent
	 * @return a set of extinguishable of the source Agent
	 */
	public static Set<SEUBuilding> getBuildingInExtinguishableRange(SEUWorldModel world, EntityID source) {
		Set<SEUBuilding> result = new FastSet<SEUBuilding>();
		double distance = world.getConfig().extinguishableDistance - EXTINGUISH_DISTANCE_THRESHOLD;
		Collection<StandardEntity> inRange = world.getObjectsInRange(source, (int)(distance * 1.5));
		for (StandardEntity entity : inRange) {
			if (entity instanceof Building) {
				if (world.getDistance(entity.getID(), source) < distance)
					result.add(world.getSeuBuilding(entity.getID()));
			}
		}
		return result;
	}
	
	public static List<EntityID> getAreaIdInExtinguishableRange(SEUWorldModel world, EntityID source) {
		List<EntityID> result = new ArrayList<EntityID>();
		double distance = world.getConfig().extinguishableDistance - EXTINGUISH_DISTANCE_THRESHOLD;
		Collection<StandardEntity> inRange = world.getObjectsInRange(source, (int)(distance * 1.5));
		
		for (StandardEntity entity : inRange) {
			if (entity instanceof Area && world.getDistance(entity.getID(), source) < distance)
				result.add(entity.getID());
		}
		return result;
	}
	
	/**
	 * Get the nearest location from a list of target location.
	 * 
	 * @param world
	 *            the world model
	 * @param locations
	 *            a set of target locations
	 * @param start
	 *            the start point
	 * @return the nearest target location
	 */
	private static EntityID getNearest(SEUWorldModel world, List<EntityID> locations, EntityID start) {
		EntityID result = null;
		double minDistance = Double.MAX_VALUE;
		for (EntityID next : locations) {
			double distance = world.getDistance(start, next);
			if (distance < minDistance) {
				minDistance = distance;
				result = next;
			}
		}
		return result;
	}
	
	public static SEUBuilding getNearest(List<SEUBuilding> buildings, Pair<Integer, Integer> source){
		SEUBuilding targetBuilding = null;
		Building building;
		Pair<Integer, Integer> buildingLocation;
		double minDistance = Double.MAX_VALUE;
		double distance;
		for (SEUBuilding next : buildings) {
			building = next.getSelfBuilding();
			buildingLocation = new Pair<Integer, Integer>(building.getX(), building.getY());
			distance = Distance.getDistance(source, buildingLocation);
			if (distance < minDistance) {
				minDistance = distance;
				targetBuilding = next;
			}
		}
		return targetBuilding;
	}
	
	/**
	 * Determines whether a target location is reachable.
	 * 
	 * @param world
	 *            the world model
	 * @param router
	 *            the router do the path plan tsk
	 * @param location
	 *            the target location
	 * @return true when the target location is reachable. Otherwise, false.
	 */
	private boolean isReacable(SEUWorldModel world, POVRouter router, EntityID location) {
		
		List<EntityID> path = router.getAStar((Human)world.getControlledEntity(), 
				(Area)world.getEntity(location), router.getStrictCostFunction());
		double timeToArrive = router.getRouteCost() / SEUWorldModel.MEAN_VELOCITY_OF_MOVING;
		double euclidDistance = world.getDistance(world.getAgent().getID(), location);
		
		if (path.size() != 0 && timeToArrive < 3 * euclidDistance) {
			return true;
		}
		
		return false;
	}
	
	public static Comparator<Pair<EntityID, Double>> pairComparator = new Comparator<Pair<EntityID, Double>>() {
		/* sorted element in increase order*/
		@Override
		public int compare(Pair<EntityID, Double> o1, Pair<EntityID, Double> o2) {
			double value1 = o1.second();
			double value2 = o2.second();
			if (value1 > value2)
				return 1;
			if (value1 < value2)
				return -1;
			return 0;
		}
	};  
	
	public static Comparator<Pair<Pair<EntityID, Double>, Double>> pairComparator_new = 
			new Comparator<Pair<Pair<EntityID, Double>, Double>>() {
		
		@Override
		public int compare(Pair<Pair<EntityID, Double>, Double> o1, Pair<Pair<EntityID, Double>, Double> o2) {
			if (o1.second().doubleValue() > o2.second().doubleValue())
				return 1;
			if (o1.second().doubleValue() < o2.second().doubleValue())
				return -1;
			
			if (o1.second().doubleValue() == o2.second().doubleValue()) {
				if (o1.first().second().doubleValue() > o2.first().second().doubleValue()) {
					return 1;
				}
				if (o1.first().second().doubleValue() < o2.first().second().doubleValue()) {
					return -1;
				}
				
				if (o1.first().second().doubleValue() == o2.first().second().doubleValue()) {
					return -1;
				}
			}
			
			return 0;
		}
	};
	
	/** Refresh the fire simulator to make the estimated result more correct.*/
	public static void refreshFireEstimator(SEUWorldModel world) {
		Building building;
		SEUBuilding seuBuilding;
		for (StandardEntity entity : world.getEntitiesOfType(AgentConstants.BUILDINGS)) {
			building = (Building) entity;
			int fireyness = building.isFierynessDefined() ? building.getFieryness() : 0;
			int temperature = building.isTemperatureDefined() ? building.getTemperature() : 0;
			
			seuBuilding = world.getSeuBuilding(entity.getID());
			if (seuBuilding.getEstimatedFieryness() > 4)
				continue;
			seuBuilding.setEnergy(temperature * seuBuilding.getCapacity(), "refreshFireEstimator");
			
			switch (fireyness) {
			case 0:
				seuBuilding.setFuel(seuBuilding.getInitialFuel());
				if (seuBuilding.getEstimatedTemperature() > seuBuilding.getIgnitionPoint())
					seuBuilding.setEnergy(seuBuilding.getIgnitionPoint() / 2, "refreshFireEstimator case 0");
				break;
			case 1:
				if (seuBuilding.getFuel() < seuBuilding.getInitialFuel() * 0.66)
					seuBuilding.setFuel(seuBuilding.getInitialFuel() * 0.75f);
				else
					seuBuilding.setFuel(seuBuilding.getInitialFuel() * 0.9f);
				break;
			case 2:
				if (seuBuilding.getFuel() < seuBuilding.getInitialFuel() * 0.33
						|| seuBuilding.getFuel() > seuBuilding.getInitialFuel() * 0.66)
					seuBuilding.setFuel(seuBuilding.getInitialFuel() * 0.5f);
				break;
			case 3:
				if (seuBuilding.getFuel() < seuBuilding.getInitialFuel() * 0.01
						|| seuBuilding.getFuel() > seuBuilding.getInitialFuel() * 0.33)
					seuBuilding.setFuel(seuBuilding.getInitialFuel() * 0.15f);
				break;
			case 8:
				seuBuilding.setFuel(0f);
				break;
			default:
				break;
			}
		}
	}

	/** Find the newest ignite building for a list of buildings.*/
	public SEUBuilding findNewestIgniteBuilding(List<SEUBuilding> buildings) {
		int minTime = Integer.MAX_VALUE;
		int tempTime;
		SEUBuilding newestIgniteBuilding = null;
		
		for (SEUBuilding next : buildings) {
			tempTime = world.getTime() - next.getIgnitionTime();
			if (tempTime < minTime) {
				minTime = tempTime;
				newestIgniteBuilding = next;
				if (minTime == 0)
					break;
			}
		}
		return newestIgniteBuilding;
	}
}
