package seu.communication;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.EnumSet;
import java.util.List;
import java.util.PriorityQueue;

import seu.communication.MessageConstant.MessageReportedType;
import seu.communication.channel.StaticAssignedChannelManager;
import seu.model.SEUWorldModel;
import seu.model.AgentConstants;
import seu.model.ConfigConstants;
import seu.util.BitArrayInputStream;
import seu.util.BitArrayOutputStream;
import seu.util.BitUtil;
import rescuecore2.messages.Command;
import rescuecore2.misc.Pair;
import rescuecore2.standard.entities.Civilian;
import rescuecore2.standard.entities.StandardEntity;
import rescuecore2.standard.messages.AKSpeak;
import rescuecore2.worldmodel.ChangeSet;


public class CommunicationManager {
	/** The world model. */
    final private SEUWorldModel world;
	/** Channel manager which do the channel assign task.*/
	final private StaticAssignedChannelManager chManager;
	
	/** radioEventListener */
	final private Port[] radioEventListener;
	private List<Port> radioEventListeners = null;
	/** The header is the index of its Port in radioEventListener.*/
	final private int radioHeaderBitSize;
	
	/** voiceEventListener */
	final private Port[] voiceEventListener;
	private List<Port> voiceEventListeners = null;
	/** The header is the index of its Port in the voiceEventListener.*/
	final private int voiceHeaderBitSize;
	
	/** civilianVoiceListener */
	final private CivilianVoiceListener[] civilianListener;
	
	/**
	 * This variable stores all possible message for agents this cycle. And it
	 * helps to handle the case that an Agent can subscribe more than one
	 * channels. First we initialize all possible message and stores them here.
	 * Then we create message for the first subscriable channel. All packet send
	 * in first channel will be removed from this queue. After the first
	 * channel, we handle the second channel, and the third and ect.
	 */
	final PriorityQueue<MessageBitSection> allMessage = new PriorityQueue<MessageBitSection>();
	
	/**
	 * When create mesage for one specified channel, we should poll all packet
	 * from {@code allMessage} and determines whether it can be put into this
	 * channel. And we should not put packet back to {@code allMessage} when it
	 * can not be put into this channel. Otherwise, it may form an infinite
	 * loop. So we should have another variable to stores those can't been put
	 * in packet. When the message creation of current channel finished,
	 * {@code allMessage} will be an empty queue, and we put all packets in this
	 * variable into {@code allMessage}. Then we can create message for next
	 * channel. Before each creation, we should clear {@code tempMessage}.
	 */
	final PriorityQueue<MessageBitSection> tempMessage = new PriorityQueue<MessageBitSection>();
	
	
	public CommunicationManager(CommunicationManagerPortsBuilder parts) {
		parts.build();
		world = parts.getWorld();                               
		chManager = parts.getChannelManager();                  
		radioEventListener = parts.getRadioEventListener();     
		voiceEventListener = parts.getVoiceEventListener();     
		radioHeaderBitSize = BitUtil.needBitSize(radioEventListener.length);   
		voiceHeaderBitSize = BitUtil.needBitSize(voiceEventListener.length);   
		civilianListener = parts.getCivilianVoiceListener();    
	}

	/**
	 * When you assign channel dynamicly, you should update channel manager in
	 * each cycle before you send a message.
	 */
	public void update() {
		chManager.update();
	}
	
	/** Get the voice communication channel. */
	public int getVoiceChannel() {
		return chManager.getSubscribeVoiceChannel();
	}

	/**
	 * Get all channels a FB Agent can subscribe to.
	 * 
	 * @return all channels a FB Agent can subscribe to
	 */
	public int[] getFireChannel(){
		return chManager.getFireChannel();
	}
	/**
	 * Get all channels a AT Agent can subscribe to.
	 * 
	 * @return all channels a AT Agent can subscribe to
	 */
	public int[] getAmbulanceChannel(){
		return chManager.getAmbulanceChannel();
	}
	/**
	 * Get all channels a PF Agent can subscribe to.
	 * 
	 * @return all channels a PF Agent can subscribe to
	 */
	public int[] getPoliceChannel(){
		return chManager.getPoliceChannel();
	}
	
	/**
	 * Get all channels this Agent can subscribe to.
	 * 
	 * @return all channels this Agent can subscribe to
	 */
	public int[] getSubscribeChannels() {
		return chManager.getSubscribeChannels();
	}

	
	
/* --------------------------- create message when all agent share one channel ----------------------------- */	
	
	
	public byte[] createRadioMessage(final int channel, final ChangeSet changed) {  
		
		final int limit = world.getConfig().radioChannels.get(channel).bandwidth;
		return createMessage(radioEventListener, radioHeaderBitSize, limit, changed, channel);
	}
	
	public byte[] createVoiceMessage(final int channel, final ChangeSet changed) {
		
		final int limit = world.getConfig().voiceChannels.get(channel).size;
		return createMessage(voiceEventListener, voiceHeaderBitSize, limit, changed, channel);
	}
	
	private byte[] createMessage(final Port[] listener, final int headerSize, 
			final int limit, final ChangeSet changed, int channel) {
		
		final PriorityQueue<MessageBitSection> que = new PriorityQueue<MessageBitSection>();
		final BitArrayOutputStream stream = new BitArrayOutputStream();
		
		for (int i = 0; i < listener.length; i++) {
			listener[i].resetCounter();
			for (listener[i].init(changed); listener[i].hasNext();) {
				MessageBitSection sec = listener[i].next();
				sec.setHeaderNumber(i);
				que.add(sec);
			}
		}
		
		final ArrayList<String> ports = new ArrayList<String>();
		
		while (!que.isEmpty()) {
			final MessageBitSection sec = que.poll();
			if ((stream.getBitLength() + sec.getBitLength()) / 8 + 1 > limit) {
				continue;
			}
			
			// TODO print the write message
			if (AgentConstants.PRINT_COMMUNICATION)  {
				listener[sec.getHeaderNumber()].printWrite(sec, channel);
			}
			// TODO
			
			ports.add(listener[sec.getHeaderNumber()].getClass().getName());
			stream.writeBit(sec.getHeaderNumber(), headerSize);
			for (Pair<Integer, Integer> writePair : sec) {
				stream.writeBit(writePair.first(), writePair.second());
			}
		}
		
		return stream.getData();
	}
	
	
	
	
/* ------------------------ create message for each agent has no more than one channel --------------------- */
	
	
	public byte[] createRadioMessage(final int channel, 
			final ChangeSet changed, EnumSet<MessageReportedType> reportedTypes) {  
		
		final int limit = world.getConfig().radioChannels.get(channel).bandwidth;
		return createMessage(radioEventListener, radioHeaderBitSize, channel, limit, changed, reportedTypes);
	}
	
	public byte[] createVoiceMessage(final int channel, 
			final ChangeSet changed, EnumSet<MessageReportedType> reportedTypes) {
		
		final int limit = world.getConfig().voiceChannels.get(channel).size;
		return createMessage(voiceEventListener, voiceHeaderBitSize, channel, limit, changed, reportedTypes);
	}
	
	private byte[] createMessage(final Port[] listener, final int headerSize, int channel,
			final int limit, final ChangeSet changed, EnumSet<MessageReportedType> reportedType) {
		
		final PriorityQueue<MessageBitSection> que = new PriorityQueue<MessageBitSection>();
		final BitArrayOutputStream stream = new BitArrayOutputStream();
		
		for (int i = 0; i < listener.length; i++) {
			Port flag = null;
			for(MessageReportedType next : reportedType) {
				if (listener[i].getMessageReportedType() == next) {
					flag = listener[i];
					break;
				}
			}
			if (flag == null)
				continue;
			listener[i].resetCounter();
			for (listener[i].init(changed); listener[i].hasNext();) {
				MessageBitSection sec = listener[i].next();
				sec.setHeaderNumber(i);
				que.add(sec);
			}
		}
		
		final ArrayList<String> ports = new ArrayList<String>();
		while (!que.isEmpty()) {
			final MessageBitSection sec = que.poll();
			if ((stream.getBitLength() + sec.getBitLength()) / 8 + 1 > limit) {
				continue;
			}	
			
			// TODO print the write message
			if (AgentConstants.PRINT_COMMUNICATION) {
				listener[sec.getHeaderNumber()].printWrite(sec, channel);
			}
			// TODO
			
			ports.add(listener[sec.getHeaderNumber()].getClass().getName());
			stream.writeBit(sec.getHeaderNumber(), headerSize);
			for (Pair<Integer, Integer> writePair : sec) {
				stream.writeBit(writePair.first(), writePair.second());
			}
		}
		
		return stream.getData();
	}
	
	
	
	
/* ---------------------- create message when each agent has many channels -------------------------------- */
	
	
	public void initRadioMessage(final ChangeSet changed, final EnumSet<MessageReportedType> reportedTypes) {
		this.initMessage(radioEventListener, changed, reportedTypes);
	}
	
	public void initVoiceMessage(final ChangeSet changed, final EnumSet<MessageReportedType> reportedTypes) {
		this.initMessage(voiceEventListener, changed, reportedTypes);
	}
	
	private void initMessage(final Port[] listener, final ChangeSet changed, 
			final EnumSet<MessageReportedType> reportedTypes) {
		
		allMessage.clear();
		for (int i = 0; i < listener.length; i++) {
			Port flag = null;
			for(MessageReportedType next : reportedTypes) {
				if (listener[i].getMessageReportedType() == next) {
					flag = listener[i];
					break;
				}
			}
			if (flag == null)
				continue;
			listener[i].resetCounter();
			for (listener[i].init(changed); listener[i].hasNext();) {
				MessageBitSection sec = listener[i].next();
				sec.setHeaderNumber(i);
				allMessage.add(sec);
			}
		}
	}
	
	public byte[] createRadioMessage(final int channel) {
		final int limit = world.getConfig().radioChannels.get(channel).bandwidth;
		return createMessage(radioHeaderBitSize, limit, radioEventListener, channel);
	}
	
	public byte[] createVoiceMessage(final int channel) {
		final int limit = world.getConfig().voiceChannels.get(channel).size;
		return createMessage(voiceHeaderBitSize, limit, voiceEventListener, channel);
	}
	
	private byte[] createMessage(final int headerSize, final int limit, final Port[] listener, int channel) {
		tempMessage.clear();
		final BitArrayOutputStream stream = new BitArrayOutputStream();
		
		final ArrayList<String> ports = new ArrayList<String>();
		while (!allMessage.isEmpty()) {
			final MessageBitSection sec = allMessage.poll();
			if ((stream.getBitLength() + sec.getBitLength()) / 8 + 1 > limit) {
				tempMessage.add(sec);
				continue;
			}
			
			// TODO print the write message
			if (AgentConstants.PRINT_COMMUNICATION) {
				listener[sec.getHeaderNumber()].printWrite(sec, channel);
			}
			// TODO
			
			ports.add(listener[sec.getHeaderNumber()].getClass().getName());
			stream.writeBit(sec.getHeaderNumber(), headerSize);
			for (Pair<Integer, Integer> writePair : sec) {
				stream.writeBit(writePair.first(), writePair.second());
			}
		}
		
		allMessage.addAll(tempMessage);
		
		return stream.getData();
	}
	
	
	
/* ----------------------------------------- read message ----------------------------------------------- */
	
	/**
	 * Read message. Including voice and radio message. This method also handle messages
	 * send by civilian.
	 * 
	 * @param heard  the received messages
	 */
	public void read(final Collection<Command> heard) {
		for (Command command : heard) {
			if (command instanceof AKSpeak) {
				final AKSpeak message = (AKSpeak) command;
				final StandardEntity sender = world.getEntity(message.getAgentID());
				final int channel = message.getChannel();

				if (sender instanceof Civilian) { 
					for (CivilianVoiceListener l : civilianListener) { 
						l.hear(message);                
					}
				} else if (chManager.isVoiceChannel(channel)) { // read voice meaasge
					readMessage(message, voiceEventListener, voiceHeaderBitSize, channel);
				} else { // read radio message
					readMessage(message, radioEventListener, radioHeaderBitSize, channel);
				}
			}
		}
	}
	
	/** Read the message send by an Agent. It can be a voice or radio message.*/
	private void readMessage(final AKSpeak message, final Port[] listener, final int headerSize, int channel) {
		
		final BitArrayInputStream stream = new BitArrayInputStream(message.getContent());
		
		final ArrayList<String> ports = new ArrayList<String>();    
		
		while (stream.hasNext()) {
			final int header = stream.readBit(headerSize);
			
			try {
				ports.add(listener[header].getClass().getName());
				listener[header].read(message.getAgentID(), message.getTime(), stream);
				
				// TODO print the readed message
				if (AgentConstants.PRINT_COMMUNICATION) {
					listener[header].printRead(channel);
				}
				// TODO
			}catch(Exception e){

			}
		}
	}
	
	public List<Port> getRadioEventListers() {
		if (radioEventListeners == null) {
			radioEventListeners = new ArrayList<Port>();
			for(Port next : radioEventListener) {
				radioEventListeners.add(next);
			}
		}
		
		return Collections.unmodifiableList(radioEventListeners);
	}
	
	public List<Port> getVoiceEventListeners() {
		if (voiceEventListeners == null) {
			voiceEventListeners = new ArrayList<Port>();
			for (Port next : voiceEventListener) {
				voiceEventListeners.add(next);
			}
		}
		return Collections.unmodifiableList(voiceEventListeners);
	}
	
	
}
