package seu.communication;

import java.util.ArrayList;
import java.util.List;

import seu.communication.channel.StaticAssignedChannelManager;
import seu.model.SEUWorldModel;
import rescuecore2.standard.entities.StandardEntity;


public class CommunicationManagerPortsBuilder {

	
	private final SEUWorldModel world;

	private final StandardEntity me;
	private final CommunicationUtil comUtil;
	private StaticAssignedChannelManager channelManager;

	/** 
	 * Stores Ports of radioEventListener. 
	 */
	private final List<Port> radioEventListener;
	/** 
	 * All Port that radio event listened. 
	 */
	private Port[] radioEventListenerResult;

	/** 
	 * Stores Ports of voiceEventListener. 
	 */
	private final List<Port> voiceEventListener;
	/** 
	 * All Port that voice event listened. 
	 */
	private Port[] voiceEventListenerResult;

	private final List<CivilianVoiceListener> civilianVoiceListener;
	private CivilianVoiceListener[] civilianListenerResult;


	public CommunicationManagerPortsBuilder(final StandardEntity me, final SEUWorldModel world) {
		this.world = world;
		this.me = me;
		this.comUtil = new CommunicationUtil(world);
		this.channelManager = new StaticAssignedChannelManager(me, world.getConfig());
		this.radioEventListener = new ArrayList<Port>();
		this.voiceEventListener = new ArrayList<Port>();
		this.civilianVoiceListener = new ArrayList<CivilianVoiceListener>();
		
		this.setWorldCommunicationCondition();
	}

	public void addRadioEventListener(Port l) {
		radioEventListener.add(l);
	}

	public void addVoiceEventListener(Port l) {
		voiceEventListener.add(l);
	}

	public void addCivilianVoiceListener(CivilianVoiceListener l) {
		civilianVoiceListener.add(l);
	}

	/**
	 * Instantiate radioEventListener, voiceEventListener, civilianVoiceListener
	 * and channelManager
	 */
	public void build() {
		int radioSize = radioEventListener.size();
		int voiceSize = voiceEventListener.size();
		int civilianSize = civilianVoiceListener.size();
		radioEventListenerResult = radioEventListener.toArray(new Port[radioSize]);
		voiceEventListenerResult = voiceEventListener.toArray(new Port[voiceSize]);
		civilianListenerResult = civilianVoiceListener.toArray(new CivilianVoiceListener[civilianSize]);
	}

	/**
	 * Set the world communication condition.
	 */
	private void setWorldCommunicationCondition() {
		CommunicationCondition condition = channelManager.getCommunicationCondition();
		switch (condition) {
		case Less:
			this.world.setCommunicationLess(true);
			break;
		case Low:
			this.world.setCommunicationLess(false);
			this.world.setCommunicationLow(true);
			break;
		case Medium:
			this.world.setCommunicationLess(false);
			this.world.setCommunicationMedium(true);
			break;
		case High:
			this.world.setCommunicationLess(false);
			this.world.setCommunicationHigh(true);
			break;
		default:
			break;
		}
	}

	SEUWorldModel getWorld() {
		return world;
	}

	StandardEntity getMe() {
		return me;
	}

	/**
	 * Return the result of radio event listened Port.
	 * 
	 * @return results of radio event listened Port
	 */
	Port[] getRadioEventListener() {
		return radioEventListenerResult;
	}

	/**
	 * Return the result of voice event listened Port.
	 * 
	 * @return results of voice event listened Port
	 */
	Port[] getVoiceEventListener() {
		return voiceEventListenerResult;
	}

	/**
	 * Return the result of civilian voice event listened Port.
	 * 
	 * @return results of civilian voice event listened Port
	 */
	CivilianVoiceListener[] getCivilianVoiceListener() {
		return civilianListenerResult;
	}

	/**
	 * Return an instance of CommunicationUtil associate with this
	 * CommunicationManagerPartsBuilder.
	 * 
	 * @return an instance of CommunicationUtil
	 */
	public CommunicationUtil getCommunicationUtil() {
		return comUtil;
	}

	/**
	 * Return an instance of ChannelManager associate with this
	 * CommunicationManagerPartsBuilder.
	 * 
	 * @return an instance of ChannelManager
	 */
	public StaticAssignedChannelManager getChannelManager() {
		return channelManager;
	}
}
